#open "unix";;
#open "obj";;
#open "meta";;
#open "signal";;
#open "misc";;
#open "config";;
#open "constants";;
#open "syntax";;
#open "builtins";;
#open "modules";;
#open "lex_run";;
#open "locations";;
#open "globals";;
#open "types";;
#open "typing";;
#open "typing_decl";;
#open "front";;
#open "back";;
#open "instruct";;
#open "opcodes";;
#open "buffcode";;
#open "labels";;
#open "symtable";;
#open "emit_code";;
#open "emit_data";;
#open "compiler";;
#open "link";;
#open "print_value";;

(* Utilisation interactive *)

exception Break;;

(* Chargement en memoire et execution d'une phrase, d'un fichier objet *)

let check_suffix name suffix =
  if sub_string name (string_length name - string_length suffix)
                     (string_length suffix) = suffix
  then name else name ^ suffix
;;

let load_phrase (init, body) =
  init_out_code();
  reset_label_table();
  literal_table := [];
  emit init;
  out STOP;
  let start = !out_pos in
  emit body;
  let stop = !out_pos in
  let code = static_alloc stop in
  blit_string !out_buffer 0 code 0 !out_pos;
  if number_of_globals() >= vect_length global_data then
    realloc_global_data(number_of_globals());
  emit_data_to_heap ();
  try
    let res = interprete code 0 in
      if start == stop then static_free code;
      res
  with x ->
    if start == stop then static_free code;
    match x with
      Break ->
        raise Break
    | _ ->
        print_begline "Uncaught exception: ";
        print_value (Repr x) type_exn;
        print_endline "";
        raise Toplevel
;;

let load_object name =
  let truename = find_in_path (check_suffix name ".zo") !path_interfaces in
  let intchan = open_in truename in
    try
      while true do
        load_phrase
          (input_value intchan : zam_instruction list * zam_instruction list)
      done
    with End_of_file ->
      close_in intchan
;;

let do_toplevel_phrase (Impl(desc,loc)) =
  begin match desc with
    Zexpr expr ->
      let ty = type_expression loc expr in
      let _ = flush std_out in
      let res = load_phrase(compile_lambda (translate_expression expr)) in
        print_begline "- = ";
        print_value res ty;
        print_endline ""
  | Zletdef(rec_flag, pat_expr_list) ->
      let env = type_letdef loc rec_flag pat_expr_list in
      let _ = flush std_out in
        if rec_flag
        then load_phrase(compile_lambda(translate_letdef_rec loc 
                                                             pat_expr_list))
        else load_phrase(compile_lambda(translate_letdef loc pat_expr_list));
        reset_rollback ();
        do_list
          (fun (name, (typ, mut_flag)) ->
            print_begline name; print_string " = ";
            print_value
              global_data.(get_slot_for_variable
                           {qual=compiled_module_name(); id=name})
              typ;
            print_endline "")
          env
  | Ztypedef decl ->
      type_typedecl loc decl
  | Zexcdef decl ->
      type_excdecl loc decl
  | Zimpldirective dir ->
      do_directive dir
  end;
  flush std_out
;;

(* Chargement d'un fichier *)

let load name =
  let truename = check_suffix name ".ml" in
  let ic = open_in truename in
  let saved_lex_inchan = !lex_inchan
  and saved_input_name = !input_name
  and saved_lex_curr_pos = !lex_curr_pos
  and saved_lex_abs_pos = !lex_abs_pos
  and saved_lex_buffer = 
    (let s = create_string lex_buffer_size in
       blit_string lex_buffer 0 s 0 lex_buffer_size; s)
  in
    init_lexer ic;
    input_name := truename;
    try
      while true do
        do_toplevel_phrase(parse_impl_phrase())
      done
    with x ->
      lex_inchan := saved_lex_inchan;
      input_name := saved_input_name;
      lex_curr_pos := saved_lex_curr_pos;
      lex_abs_pos := saved_lex_abs_pos;
      blit_string saved_lex_buffer 0 lex_buffer 0 lex_buffer_size;
      match x with
        End_of_file -> ()
      | Toplevel -> ()
      | _ -> raise x
;;

(* Pour sortir --- bien que "^D" marche aussi bien... *)

let quit x = exit 0
;;

(* Cette saloperie de trace *)

let (trace, untrace) =
  let trace_env = ref [] in
  let rec trace_instr name obj ty =
    match type_repr ty with
      Typ(Tarrow(t1,t2), _) ->
        Repr(fun arg ->
          print_begline name; print_string " <-- ";
          print_value arg t1; print_endline "";
          try
            let res = (magic_obj obj : obj -> obj) arg in
              print_begline name; print_string " --> ";
              print_value res t2; print_endline "";
              trace_instr (name ^ "*") res t2
          with exc ->
            print_begline name; print_string " raises exception ";
            print_value (Repr exc) type_exn; print_endline "";
            raise exc)
    | _ -> obj in
  (* trace *)
  (fun name ->
    try
      let val_desc = find_value_desc (GRname name) in
      let pos = get_slot_for_variable val_desc.qualid in
        if mem_assoc pos !trace_env then begin
          prerr_begline "> "; prerr_string name;
          prerr_endline " is already traced."
        end else begin
          trace_env := (pos, global_data.(pos)) :: !trace_env;
          global_data.(pos) <-
            trace_instr name global_data.(pos) val_desc.info.val_typ;
          prerr_begline "> "; prerr_string name;
          prerr_endline " is now traced."
        end
    with failure _ ->
      prerr_begline ">> "; prerr_string name; prerr_endline " is undefined."),
  (* untrace *)
  (fun name ->
    try
      let val_desc = find_value_desc (GRname name) in
      let pos = get_slot_for_variable val_desc.qualid in
      let rec except = function
        [] -> prerr_begline "> "; prerr_string name;
              prerr_endline " was not traced.";
              []
      | (pos',obj as pair)::rest ->
          if pos == pos' then begin
            global_data.(pos) <- obj;
            prerr_begline "> "; prerr_string name;
            prerr_endline " is no more traced.";
            rest
          end else
            pair :: except rest
      in
        trace_env := except !trace_env;
        ()
    with failure _ ->
      prerr_begline ">> "; prerr_string name; prerr_endline " is undefined.")
;;

(* Pour definir ses propres fonctions d'impression *)

let new_printer name printer =
  try
    printers := (find_type_desc (GRname name), magic printer) :: !printers;
    ()
  with failure _ ->
    prerr_begline ">> Type "; prerr_string name;
    prerr_endline " is undefined."
;;

let default_printer name =
  try
    let ty_constr = find_type_desc (GRname name) in
    let rec remove = function
      [] -> []
    | (ty,_ as pair)::rest ->
        if same_type_constr ty ty_constr then rest else pair :: remove rest in
    printers := remove !printers;
    ()
  with failure _ ->
    prerr_begline ">> Type "; prerr_string name;
    prerr_endline " is undefined."
;;


(* La boucle principale *)

toplevel := true;;
prerr_begline version__banner;
prerr_endline2 "";
path_interfaces := [config__path_library; misc__current_dir_name];
default_used_modules := "toplvl" :: assoc "cautious" default_used_interfaces;
try
  let current_arg = ref 1 in
  let arg_of_option () =
    if incr current_arg < vect_length command_line
    then command_line.(!current_arg)
    else begin
      prerr_begline ">> Missing parameter to option ";
      prerr_endline command_line.(pred !current_arg);
      raise Toplevel
    end in
  while !current_arg < vect_length command_line do
    begin match command_line.(!current_arg) with
      "-I" -> path_interfaces := arg_of_option() :: !path_interfaces; ()
    | "-O" -> begin let set = arg_of_option() in
                try
                  default_used_modules :=
                    assoc set default_used_interfaces; ()
                with failure "find" ->
                  prerr_begline ">> Unknown module set: ";
                  prerr_endline2 set
              end
    |  s   -> prerr_begline "> Unrecognized option ";
              prerr_endline2 s
    end;
    incr current_arg
  done;
  let ic = open_in command_line.(0) in
    while input_char ic != `\010` do () done;
    let size_code = input_int ic in
    let size_data = input_int ic in
    let size_symb = input_int ic in
    let size_debug = input_int ic in
      seek_in ic (pos_in ic + size_code + size_data);
      load_linker_tables ic;
      reset_c_primitives();
      close_in ic;
      new_globalenv "top";
      init_lexer std_in;
      set_signal SIGINT (Signal_raise Break);
      while true do
        try
          print_string config__toplevel_input_prompt;
          flush std_out;
          do_toplevel_phrase(parse_impl_phrase())
        with End_of_file ->
               close_out std_out; close_out std_err; exit 0
           | Toplevel ->
               flush std_out
           | Break ->
               print_begline "Interrupted."; print_endline "";
               flush std_out;
               protect rollback ()
      done
with x ->
  prerr_begline ">> Uncaught internal exception.";
  prerr_endline "";
  close_out std_out; close_out std_err; exit 4
;;
